use axum::extract::{Path, State};
use axum::http::header::SET_COOKIE;
use axum::http::StatusCode;
use axum::response::{AppendHeaders, Html, IntoResponse, Redirect, Response};
use axum::Form;
use axum::{
    routing::{get, post},
    Json, Router,
};
use axum_extra::extract::CookieJar;
use serde::{Deserialize, Serialize};
use std::net::SocketAddr;
use std::sync::{Arc, Mutex};

#[derive(Debug, Clone, Serialize, Deserialize)]
struct Page {
    slug: String,
    votes: i32,
}

#[derive(Debug, Clone, Serialize, Deserialize, Default)]
struct VoteRecord {
    pages: Vec<Page>,
}

#[derive(Debug, Clone, Serialize, Deserialize, PartialEq)]
struct PollChoice {
    id: i32,
    votes: i32,
}

#[derive(Debug, Clone, Serialize, Deserialize, PartialEq)]
struct Poll {
    id: i32,
    choices: Vec<PollChoice>,
}

#[derive(Debug, Clone, Serialize, Deserialize, Default)]
struct PollRecord {
    polls: Vec<Poll>,
}

#[derive(Clone)]
struct AppState {
    record: Arc<Mutex<VoteRecord>>,
    poll_votes: Arc<Mutex<PollRecord>>,
}

impl AppState {
    fn save(&self) {
        self.save_polls();
        self.save_votes();
    }

    fn save_polls(&self) {
        let st = self.poll_votes.lock().unwrap().to_owned();
        serde_json::to_writer(&std::fs::File::create("pollvotes.json").unwrap(), &st)
            .expect("failed to write poll votes!");
    }

    fn save_votes(&self) {
        let st = self.record.lock().unwrap().to_owned();
        serde_json::to_writer(&std::fs::File::create("votes.json").unwrap(), &st)
            .expect("failed to write votes!");
    }

    fn vote_poll(&self, poll_id: i32, choice_id: i32) {
        let mut poll_votes = self.poll_votes.lock().unwrap();

        // Ensure we even have a valid poll id
        let Some(poll_config) = AppState::get_poll_config_by_id(poll_id) else {
            return;
        };

        let mut poll: Option<&mut Poll> = None;
        for existing_poll in &mut poll_votes.polls {
            if existing_poll.id == poll_id {
                poll = Some(existing_poll);
                break;
            }
        }
        if poll.is_none() {
            let mut choices = vec![];
            for choice in poll_config.choices {
                choices.push(PollChoice {
                    id: choice.id,
                    votes: 0,
                });
            }

            poll_votes.polls.push(Poll {
                id: poll_id,
                choices,
            });
            poll = poll_votes.polls.last_mut();
        }
        let poll = poll.unwrap();

        for choice in &mut poll.choices {
            if choice.id == choice_id {
                choice.votes += 1;
                break;
            }
        }
    }

    fn get_poll_config_by_id(id: i32) -> Option<PollConfig> {
        if let Ok(data) = std::fs::read_to_string("polls.json") {
            if let Ok(config) = serde_json::from_str::<PollsConfig>(&data) {
                let current_poll: Vec<&PollConfig> =
                    config.polls.iter().filter(|poll| poll.id == id).collect();
                if !current_poll.is_empty() {
                    return Some(<&PollConfig>::clone(current_poll.first().unwrap()).clone());
                }
            }
        }

        None
    }

    fn get_poll_votes_by_poll_id(&self, poll_id: i32) -> Option<Poll> {
        let poll_votes = self.poll_votes.lock().unwrap();

        for existing_poll in &poll_votes.polls {
            if existing_poll.id == poll_id {
                return Some(existing_poll.clone());
            }
        }

        None
    }
}

async fn view_votes(State(state): State<AppState>, Path(slug): Path<String>) -> Json<Page> {
    tracing::info!("Requesting votes for {slug}");

    for page in &state.record.lock().unwrap().pages {
        if page.slug == slug {
            return Json(page.clone());
        }
    }

    Json(Page { slug, votes: 0 })
}

async fn leaderboard_submit_score(State(state): State<AppState>, Path(slug): Path<String>) {
    tracing::info!("Submitting vote for {slug}");

    let mut found = false;
    for page in &mut state.record.lock().unwrap().pages {
        if page.slug == slug {
            page.votes += 1;
            found = true;
        }
    }

    if !found {
        state
            .record
            .lock()
            .unwrap()
            .pages
            .push(Page { slug, votes: 1 });
    }

    state.save_votes();
}

#[derive(Debug, Clone, Deserialize)]
struct PollChoiceConfig {
    id: i32,
    name: String,
    url: String,
}

#[derive(Debug, Clone, Deserialize)]
struct PollConfig {
    id: i32,
    title: String,
    choices: Vec<PollChoiceConfig>,
}

#[derive(Debug, Clone, Deserialize)]
struct PollsConfig {
    current_poll: Option<i32>,
    polls: Vec<PollConfig>,
}

#[derive(Debug, Clone, Serialize)]
struct PollChoiceResponse {
    id: i32,
    name: String,
    url: String,
    votes: i32,
}

#[derive(Deserialize)]
struct SubmitForm {
    choice: i32,
}

async fn vote_current_poll(
    State(state): State<AppState>,
    Form(form): Form<SubmitForm>,
) -> Response {
    let id = form.choice;
    tracing::info!("Submitting vote for choice {id}");

    if let Ok(data) = std::fs::read_to_string("polls.json") {
        if let Ok(config) = serde_json::from_str::<PollsConfig>(&data) {
            if let Some(current_poll) = config.current_poll {
                let current_poll: Vec<&PollConfig> = config
                    .polls
                    .iter()
                    .filter(|poll| poll.id == current_poll)
                    .collect();
                if !current_poll.is_empty() {
                    state.vote_poll(current_poll.first().unwrap().id, id);
                    state.save_polls();

                    let headers = AppendHeaders([(
                        SET_COOKIE,
                        format!("poll_{}={}", current_poll.first().unwrap().id, id),
                    )]);

                    // make sure to redirect the iframe back to results
                    return (headers, Redirect::to("/polls/frame/results")).into_response();
                }
            }
        }
    }

    (StatusCode::NOT_FOUND, "No poll is open right now!").into_response()
}

async fn poll_frame_vote(State(state): State<AppState>, jar: CookieJar) -> Html<String> {
    tracing::info!("Requesting current poll");
    if let Ok(data) = std::fs::read_to_string("polls.json") {
        if let Ok(config) = serde_json::from_str::<PollsConfig>(&data) {
            if let Some(current_poll) = config.current_poll {
                let current_poll: Vec<&PollConfig> = config
                    .polls
                    .iter()
                    .filter(|poll| poll.id == current_poll)
                    .collect();
                if !current_poll.is_empty() {
                    let poll_config = current_poll.first().unwrap();
                    let mut choices = vec![];

                    let find_vote_choice = |id: i32| {
                        if let Some(poll) = state.get_poll_votes_by_poll_id(current_poll[0].id) {
                            for choice in &poll.choices {
                                if choice.id == id {
                                    return Some(choice.clone());
                                }
                            }
                        }

                        None
                    };

                    for choice in &poll_config.choices {
                        if let Some(vote_choice) = find_vote_choice(choice.id) {
                            choices.push(PollChoiceResponse {
                                id: choice.id,
                                name: choice.name.clone(),
                                url: choice.url.clone(),
                                votes: vote_choice.votes,
                            });
                        } else {
                            choices.push(PollChoiceResponse {
                                id: choice.id,
                                name: choice.name.clone(),
                                url: choice.url.clone(),
                                votes: 0,
                            });
                        }
                    }

                    let mut html = String::new();
                    html.push_str(&format!(
                        "<h3>{}</h3>",
                        current_poll.first().unwrap().title.clone()
                    ));

                    let mut existing_choice = None;
                    for cookie in jar.iter() {
                        if cookie.name().contains("poll_") {
                            let id = cookie.name().replace("poll_", "");
                            if let Ok(id) = id.parse::<i32>() {
                                if id == poll_config.id {
                                    if let Ok(value) = cookie.value().parse::<i32>() {
                                        existing_choice = Some(value);
                                    }
                                }
                            }
                        }
                    }

                    html.push_str("<form id=\"results\" action=\"/polls/frame/results\">");
                    html.push_str("</form>");

                    html.push_str("<form action=\"/polls/vote\" method=\"post\">");
                    html.push_str("<fieldset>");
                    let inert = if existing_choice.is_some() {
                        "inert"
                    } else {
                        ""
                    };
                    html.push_str(&format!("<div {}>", inert));
                    for choice in choices {
                        let checked = if Some(choice.id) == existing_choice {
                            "checked"
                        } else {
                            ""
                        };
                        html.push_str(&format!("<input type=\"radio\" name=\"choice\" id=\"choice{}\" value=\"{}\" {}>", choice.id, choice.id, checked));
                        html.push_str(&format!(
                            "<label for=\"choice{}\">{}</label>",
                            choice.id,
                            choice.name.clone()
                        ));
                    }
                    html.push_str("</div>");
                    html.push_str(&format!("<input type=\"submit\" {}/>", inert));
                    html.push_str(
                        "<input form=\"results\" type=\"submit\" value=\"View Results\"/>",
                    );
                    html.push_str("</fieldset>");
                    html.push_str("</form>");

                    return Html(html);
                }
            }
        }
    }

    Html("<html><p>No poll is open right now!</p></html>".to_string())
}

async fn poll_frame_results(State(state): State<AppState>, jar: CookieJar) -> Html<String> {
    if let Ok(data) = std::fs::read_to_string("polls.json") {
        if let Ok(config) = serde_json::from_str::<PollsConfig>(&data) {
            if let Some(current_poll) = config.current_poll {
                let current_poll: Vec<&PollConfig> = config
                    .polls
                    .iter()
                    .filter(|poll| poll.id == current_poll)
                    .collect();
                if !current_poll.is_empty() {
                    let poll_config = current_poll.first().unwrap();
                    let mut choices = vec![];

                    let find_vote_choice = |id: i32| {
                        if let Some(poll) = state.get_poll_votes_by_poll_id(current_poll[0].id) {
                            for choice in &poll.choices {
                                if choice.id == id {
                                    return Some(choice.clone());
                                }
                            }
                        }

                        None
                    };

                    for choice in &poll_config.choices {
                        if let Some(vote_choice) = find_vote_choice(choice.id) {
                            choices.push(PollChoiceResponse {
                                id: choice.id,
                                name: choice.name.clone(),
                                url: choice.url.clone(),
                                votes: vote_choice.votes,
                            });
                        } else {
                            choices.push(PollChoiceResponse {
                                id: choice.id,
                                name: choice.name.clone(),
                                url: choice.url.clone(),
                                votes: 0,
                            });
                        }
                    }

                    let mut html = String::new();
                    html.push_str(&format!(
                        "<h3>{}</h3>",
                        current_poll.first().unwrap().title.clone()
                    ));

                    html.push_str("<form id=\"results\" action=\"/polls/frame/vote\">");
                    html.push_str("</form>");

                    let mut existing_choice = None;
                    for cookie in jar.iter() {
                        if cookie.name().contains("poll_") {
                            let id = cookie.name().replace("poll_", "");
                            if let Ok(id) = id.parse::<i32>() {
                                if id == poll_config.id {
                                    if let Ok(value) = cookie.value().parse::<i32>() {
                                        existing_choice = Some(value);
                                    }
                                }
                            }
                        }
                    }

                    html.push_str("<form>");
                    html.push_str("<fieldset>");
                    html.push_str("<div inert>");
                    for choice in choices {
                        let checked = if Some(choice.id) == existing_choice {
                            "checked"
                        } else {
                            ""
                        };

                        html.push_str(&format!("<input type=\"radio\" name=\"choice\" id=\"choice{}\" value=\"{}\" {}>", choice.id, choice.id, checked));
                        html.push_str(&format!(
                            "<label for=\"choice{}\">{} ({} votes)</label>",
                            choice.id,
                            choice.name.clone(),
                            choice.votes
                        ));
                    }
                    html.push_str("</div>");
                    html.push_str("<input form=\"results\" type=\"submit\" value=\"Back\"/>");
                    html.push_str("</fieldset>");
                    html.push_str("</form>");

                    return Html(html);
                }
            }
        }
    }

    Html("<html><p>No poll is open right now!</p></html>".to_string())
}

#[tokio::main]
async fn main() {
    tracing_subscriber::fmt::init();

    let vote_record = if let Ok(data) = std::fs::read_to_string("votes.json") {
        serde_json::from_str(&data).expect("Failed to parse votes data")
    } else {
        VoteRecord::default()
    };

    let poll_record = if let Ok(data) = std::fs::read_to_string("pollvotes.json") {
        serde_json::from_str(&data).expect("Failed to parse polls data")
    } else {
        PollRecord::default()
    };

    let initial_state = AppState {
        record: Arc::new(Mutex::new(vote_record)),
        poll_votes: Arc::new(Mutex::new(poll_record)),
    };

    let app = Router::new()
        .route("/votes/{slug}", post(leaderboard_submit_score))
        .route("/votes/{slug}", get(view_votes))
        .route("/polls/vote", post(vote_current_poll))
        .route("/polls/frame/vote", get(poll_frame_vote))
        .route("/polls/frame/results", get(poll_frame_results))
        .with_state(initial_state.clone());

    let addr = SocketAddr::from(([127, 0, 0, 1], 9183));
    tracing::info!("Listening on {}", addr);
    let listener = tokio::net::TcpListener::bind(&addr).await.unwrap();
    axum::serve(listener, app).await.unwrap();

    let state_clone = initial_state.clone();
    state_clone.save();
}
